var/list/html_entities = list("&#34;" = "\"","&#39;" = "\'")

client/New()
	.=..()
	src << browse_rsc('dm_syntax.css')

proc/syntax_highlight(code,full_doc = 0,title = "DM Code")
	ASSERT(istext(code))
	//censor HTML
	code = html_encode(code)
	for(var/entity in html_entities)
		code = _dmsh_string_replace(code,entity,html_entities["[entity]"])
	var/in_singstring   = 0
	var/in_doubstring   = 0
	var/in_bigstring    = 0
	var/in_brackets     = 0
	var/in_singcomment  = 0
	var/in_multicomment = 0
	var/in_preprocess   = 0
	//Don't highlight quotes around HTML attributes
	var/in_tag = 0
	//Detect if any changes have been made during last loop iteration
	var/change_made      = 0
	var/last_change_made = 0
	var/list/found_keywords
	var/pos = 0
	while(1)
		pos++
		last_change_made = change_made
		change_made = 0
		//The text character currently being looked at
		var/char = text2ascii(code,pos)
		//if unescaped newline is starting
		if(char == 10 && pos > 2 && copytext(code,pos - 2,pos - 1) != 92)
			if(in_preprocess)
				code = _insert_text(code,"</span>",pos + 1)
				in_preprocess = 0
				pos += length("</span>")
				change_made = 1
			if(in_singcomment)
				code = _insert_text(code,"</span>",pos + 1)
				in_singcomment = 0
				pos += length("</span>")
				change_made = 1
		//if multicomment is ending
		if(char == 47 && pos > 1 && text2ascii(code,pos - 1) == 42)
			code = _insert_text(code,"</span>",pos + 1)
			in_multicomment = 0
			pos += length("</span>")
			change_made = 1
		//if we aren't inside a comment and the char isn't escaped
		if(!in_singcomment && !in_multicomment)
			//escaped text
			if(pos > 1 && text2ascii(code,pos - 1) == 92 && (in_doubstring || in_bigstring))
				//account for an escaped backslash
				if(pos > 2)
					if(!(text2ascii(code,pos - 1) == 92 && text2ascii(code,pos - 2) == 92)) continue
				else continue
			//singstring
			if(char == 39 && !in_doubstring && !in_tag)
				if(in_singstring && !in_brackets && !in_bigstring)
					code = _insert_text(code,"</span>",pos + 1)
					in_singstring = 0
					pos += length("</span>")
					change_made = 1
				else
					code = _insert_text(code,"<span class=\"dmstring\">",pos)
					in_singstring = 1
					pos += length("<span class=\"dmstring\">")
					change_made = 1
			//doubstring start + end and bigstring start
			if(char == 34)
				//doubstring end
				if(in_doubstring && !in_brackets && !in_bigstring && !in_tag)
					code = _insert_text(code,"</span>",pos + 1)
					in_doubstring = 0
					pos += length("</span>") + 1
					change_made = 1
				else
					//bigstring start
					if(pos > 1 && text2ascii(code,pos - 1) == 123 && !in_bigstring)
						code = _insert_text(code,"<span class=\"dmstring\">",pos - 1)
						in_bigstring = 1
						pos += length("<span class=\"dmstring\">") + 1
						change_made = 1
					//doubstring start
					else if(!in_doubstring && !in_bigstring && !in_singstring && !in_tag)
						code = _insert_text(code,"<span class=\"dmstring\">",pos)
						in_doubstring = 1
						pos += length("<span class=\"dmstring\">")
						change_made = 1
			//bigstring end
			if(char == 125 && in_bigstring && text2ascii(code,pos - 1) == 34)
				code = _insert_text(code,"</span>",pos + 1)
				in_bigstring = 0
				pos += length("</span>") + 1
				change_made = 1
			if(!in_doubstring && !in_singstring && !in_bigstring)
				//Check for comment starts
				if(pos > 1 && text2ascii(code,pos - 1) == 47)
					if(char == 47)
						//singcomment after preprocessor -- on same line
						if(in_preprocess)
							code = _insert_text(code,"</span>",pos - 1)
							in_preprocess = 0
							pos += length("</span>")
						code = _insert_text(code,"<span class=\"dmcomment\">",pos - 1)
						in_singcomment = 1
						pos += length("<span class=\"dmcomment\">")
						change_made = 1
					if(char == 42)
						code = _insert_text(code,"<span class=\"dmcomment\">",pos - 1)
						in_multicomment = 1
						pos += length("<span class=\"dmcomment\">") + 1//@SPECIAL
						change_made = 1
				//preprocessor
				if(char == 35)
					code = _insert_text(code,"<span class=\"dmpreprocessor\">",pos)
					in_preprocess = 1
					pos += length("<span class=\"dmpreprocessor\">")
					change_made = 1
				//keywords
				if(last_change_made || pos == 1) found_keywords = get_dm_keywords(code)
				if((num2text(pos) in found_keywords) && !in_preprocess)
					code = _insert_text(code,"<span class=\"dmkeyword\">",pos)
					pos += length("<span class=\"dmkeyword\">") + length(found_keywords["[pos]"])
					code = _insert_text(code,"</span>",pos)//@SPECIAL
					pos += length("</span>") - 1
					change_made = 1
			if(in_doubstring)
				//in_tag
				if(char == 60)           in_tag = 1
				if(char == 62 && in_tag) in_tag = 0
			if(in_doubstring || in_bigstring)
				//bracket start
				if(char == 91)
					in_brackets++
					if(in_brackets == 1)
						//end string
						code = _insert_text(code,"</span>",pos)
						pos += length("</span>")
						//start brackets
						code = _insert_text(code,"<span class=\"dmbrace\">",pos)
						pos += length("<span class=\"dmbrace\">")
						change_made = 1
				//bracket end
				if(char == 93 && in_brackets)
					in_brackets--
					if(!in_brackets)
						code = _insert_text(code,"</span>",pos + 1)
						pos += length("</span>") + 1
						//continue on with string
						code = _insert_text(code,"<span class=\"dmstring\">",pos)
						pos += length("<span class=\"dmstring\">")
						//hacky doubstring stop
						if(in_doubstring && text2ascii(code,pos) == 34)
							code = _insert_text(code,"</span>",pos + 1)
							in_doubstring = 0
							pos += length("</span>") + 1
						change_made = 1
		//Break out of the loop here
		if(pos >= length(code) || char == 0)
			if(!full_doc) return "<pre class=\"dmcode\">[code]</pre>"
			else return {"
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
  <head>
    <link rel="stylesheet" type="text/css" href="dm_syntax.css" />
    <title>[title]</title>
  </head>
  <body>
<pre class=\"dmcode\">[code]</pre>
  </body>
</html>
"}